#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <values.h>
#include <time.h>

#define BLOCKSIZE 4*1024*1024

#define MAXSTR 256

typedef enum {NONE, CHECK_1PPS, CHECK_COUNT, CHECK_STATS, 
	      SEARCH_PATTERN} mode_type;
typedef enum {VLBA, AT} encoding_type;

double read_head(int file, struct tm *tm);

int main (int argc, char * const argv[]) {
  int nfile, i, j, file, status, nread, memsize, opt, first, ppsOK;
  int secount, bytes, tmp, psize, poverlap, headsize, mhz64;
  unsigned int *data, diff, last;
  unsigned short *start;
  unsigned long *stats;
  char msg[MAXSTR+1], searchstr[MAXSTR+1];
  unsigned char  *ptr, *pattern, *mem;
  double time;
  off_t off, offset;
  struct tm tm;

  mode_type mode;   /* Run mode   CHECK_1PPS:  Look for 0xff[ff] each second
                                  CHECK_COUNT: Check consistency of data
				               from VSIC countdown mode */
  int verbose;      /* Prints whats happening */
  int fail;         /* Don't keep going after a failure */
  int bits;         /* Total number of bits/sample */
  int rate;         /* Recording rate in Msamples/sec (date rate is rate*bits) */
  int maxbytes;     /* Maximum number of bytes to read in. Default all */
  encoding_type encoding; /* VLBA or AT bit encoding (offset binary or sign/mag) */

  struct option options[] = {
    {"pps", 0, 0, 'p'},
    {"count", 0, 0, 'c'},
    {"stats", 0, 0, 's'},
    {"bits", 1, 0, 'b'},
    {"rate", 1, 0, 'r'},
    {"search", 1, 0, 'S'},
    {"nbytes", 1, 0, 'n'},
    {"vlba", 0, 0, 'v'},
    {"wide", 0, 0, 'w'},
    {"64", 0, 0, 'w'},
    {"fail", 0, 0, 'f'},
    {"verbose", 0, 0, 'V'},
    {"help", 0, 0, 'h'},
    {0, 0, 0, 0}
  };

  mode = NONE;
  verbose = 0;
  fail = 0;
  stats = NULL;
  bits = 8;
  rate = 32;
  maxbytes = -1;
  encoding = AT;
  mhz64 = 0;

  /* Keep compiler happy */
  psize = 0;
  pattern = NULL;
  poverlap = 0;
  last = 0;
  start = 0;

  /* Read command line options */
  while (1) {
    opt = getopt_long_only(argc, argv, "", options, NULL);
    if (opt==EOF) break;

    switch (opt) {
      
    case 'p':
      if (mode!=NONE) {
	fprintf(stderr, "Multiple modes requested aborting\n");
	return(1);
      }
      mode = CHECK_1PPS;
      break;

    case 'c':
      if (mode!=NONE) {
	fprintf(stderr, "Multiple modes requested aborting\n");
	return(1);
      }
      mode = CHECK_COUNT;
      break;

    case 's':
      if (mode!=NONE) {
	fprintf(stderr, "Multiple modes requested aborting\n");
	return(1);
      }
      mode = CHECK_STATS;
      break;

    case 'b':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -bits option %s\n", optarg);
      else 
	bits = tmp;
      break;

    case 'r':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -rate option %s\n", optarg);
      else 
	rate = tmp;
      break;

    case 'S':
      if (mode!=NONE) {
	fprintf(stderr, "Multiple modes requested aborting\n");
	return(1);
      }
      if (strlen(optarg)>MAXSTR) {
	fprintf(stderr, "Search pattern too long (%d chars).\n", 
		strlen(optarg));
	exit(1);
      }
      strcpy(searchstr, optarg);
      mode = SEARCH_PATTERN;
      break;

    case 'n':
      status = sscanf(optarg, "%d", &tmp);
      if (status!=1)
 	fprintf(stderr, "Bad -nbytes option %s\n", optarg);
      else 
	maxbytes = tmp;
      break;

    case 'V':
      verbose = 1;
      break;

    case 'f':
      fail = 1;
      break;

    case 'v':
      encoding = VLBA;
      break;

    case 'w':
      mhz64 = 1;
      break;

    case 'h':
      printf("Usage: vsib_checker [options]\n");
      printf("  -p/-pps          Check 1 PPS markers (default)\n");
      printf("  -c/-count        Check VSIC countdown mode\n");
      printf("  -s/-stats        Check sampler stats\n");
      printf("  -S/-search <hex> Search for bit pattern in data "
	     "(specified as hexnumber\n");
      printf("  -b/-bits <b>     Total number bits/sample (8/16/32)\n");
      printf("  -r/-rate <r>     Recoding rate Msamples/sec (default 32)\n");
      printf("  -n/-nbytes <n>   Maximum number of bytes to read per file\n");
      printf("  -v/-vlba         Calculate statistics for VLBA data\n");
      printf("  -w/-wide/-64     Assume 64 MHz bandwidth data\n");
      printf("  -f/-fail         Stop runnign when failure found\n");
      printf("  -V/-verbose      Verbose\n");
      printf("  -h/-help         This list\n");
      return(1);
      break;

    case '?':
    default:
      break;
    }
  }
  
  if (mode==NONE) {
    mode = CHECK_1PPS;
    printf("Running in 1 PPS checking mode\n");
  }

  bytes = bits/8;

  if (mode==CHECK_COUNT || mode==CHECK_STATS || mode==SEARCH_PATTERN) {
    memsize = BLOCKSIZE;
  } else if (mode==CHECK_1PPS) {
    memsize = bytes;
  } else {
    fprintf(stderr, "Internal error. Unknown mode\n");
    return(1);
  }

  mem = malloc(memsize);
  if (mem==NULL) {
    sprintf(msg, "Trying to allocate %d byes", memsize);
    perror(msg);
    return(1);
  }

  if (mode==CHECK_STATS) {
    stats = malloc(256*sizeof(long)*bytes);
    if (stats==NULL) {
      sprintf(msg, "Trying to allocate %d byes", 256*sizeof(long)*bytes);
      perror(msg);
      return(1);
    }
  } else if (mode==SEARCH_PATTERN) {
    char *cptr, *pptr, str[3];
    /* Convert the search string into a number */
    /* Assume the number is hex */
    psize = floor(strlen(searchstr)/2);
    pattern = malloc(psize);
    if (pattern==NULL) {
      sprintf(msg, "Trying to allocate %d byes", psize);
      perror(msg);
      return(1);
    }

    /* Assume no leading 0x or anything */
    cptr = searchstr;
    pptr = pattern;
    for (i=0; i<psize; i++) {
      strncpy(str, cptr, 2);
      if (!isxdigit(str[0]) || !isxdigit(str[1])) {
	fprintf(stderr, "Invalid digit %s in pattern\n", str);
	return(1);
      }
      status = sscanf(str, "%x", &tmp);
      if (status!=1) {
	fprintf(stderr, "Error converting %s into pattern byte\n", str);
	return(1);
      } else if (tmp>256) {
	fprintf(stderr, "Error converting %s into pattern "
		"- value too large\n", str);
	return(1);
      }
      *pptr = (unsigned char)tmp;

      cptr +=2;
      pptr++;
      
    }
  }

  for (nfile=optind; nfile<argc; nfile++) {
    /* Initialise stuff */
    if (mode==CHECK_STATS) {
      memset(stats, 0, 256*sizeof(unsigned long)*bytes);
    }

    file = open(argv[nfile], O_RDONLY|O_LARGEFILE);
    if (file==-1) {
      sprintf(msg, "Failed to open input file (%s)", argv[nfile]);
      perror(msg);
      free(mem);
      return(1);
    }
    printf("%s\n", argv[nfile]);


    /* Read the header */
    time = read_head(file, &tm);

    headsize = lseek(file, 0, SEEK_CUR); /* Get the header size */

    offset = 0;
    first = 1;
    secount = 1;
    while (1) {
      if (mode==CHECK_1PPS) {      
	if (!first) {
	  /* Skip to next 1 PPS marker */
	  off = lseek(file, rate*1e6*bytes-bytes, SEEK_CUR); 
	  offset += rate*1e6*bytes;
	  if (off==-1) {
	    sprintf(msg, "\nFailed seek to next second");
	    perror(msg);
	    close(file);
	    return(1);
	  } else if (off != offset+headsize) {
	    fprintf(stderr, "\nFailed seek to next second 2\n");
	    close(file);
	    return(1);
	  }
	} else {
	  first = 0;
	}
      }

      nread = read(file, mem, memsize);
      if (verbose && mode==CHECK_1PPS) printf("Read second %d\n", secount);
      secount++;

      if (nread==0) {
	if (offset==0) {
	  fprintf(stderr, "No data read from %s\n", argv[nfile]);
	  free(mem);
	  close(file);
	  return(1);
	}
	break;
      } else if (nread==-1) {
	perror("Error reading file");
	close(file);
	free(mem);
	return(1);
      } else if (mode==CHECK_COUNT && nread%4) {
	/* Need to read in multiple of 4 bytes */
	off = lseek(file, -(nread%8), SEEK_CUR); 
	if (off==-1) {
	  perror(msg);
	  free(mem);
	  return(1);
	}
	nread -= nread%4;
      }

      if (mode==CHECK_COUNT) {
	unsigned int expected, wrap1, wrap2, pps, ppssize;
	ppssize = rate*bytes*1e6;

	switch (bytes) {
	case 1:
	  expected = 67372036;
	  wrap1 = 0x10203;
	  wrap2 = 0xfcfdfeff ;
	  pps = 0xff;
	  break;
	case 2:
	  expected = 131074;
	  wrap1 = 0x1;
	  wrap2 = 0xfffeffff;
	  pps = 0xffff;
	  break;
	case 4:
	  expected = 1; 
	  wrap1 = 0x0;
	  wrap2 = 0x0;
	  pps = 0xffffffff;
	  break;
	default:
	  expected = 0;
	  wrap1 = 0x0;
	  wrap2 = 0x0;
	  pps = 0x0;
	  printf("Do not support bit mod %d\n", bits);
	  break;
	}

	data = (unsigned int*)mem;
	for (i=0; i<nread/4; i++) {
	  if (first) {
	    first = 0;
	    last = *data;
	    data++;
	    continue;
	  }

	  //start = (unsigned short*) data;
	  diff = last - *data;

	  if (diff != expected) { /* Something is wrong */
	    int OK = 0;
	    if (last==wrap1 && *data==wrap2) OK = 1; // Counter wrapped
	    if (offset%ppssize==0 &&  (last&pps)==pps) OK =1; /* Ignore 1 PPS */
	    if (offset%ppssize==(ppssize-4) &&  (*data&pps)==pps) OK =1;

	    if (!OK) {
	      printf("Bad values at offset %ld\n", offset);
	      printf("     0x%04x 0x%04x\n", last, *data);
	      //printf("     0x%04x 0x%04x 0x%04x 0x%04x\n", start[0], start[1], 
	      //     start[2], start[3]);
	      printf("   %d\n", diff);
	      if (fail) return(2);
	    }
	  }
	  last = *data;
	  data++;
	  offset += 4;
	}
      } else if (mode==CHECK_1PPS) {
	ppsOK = 1;
	if (bytes==1) {
	  if ((unsigned char)*mem!=0xFF)
	    ppsOK = 0;
	    sprintf(msg, "0x%02x", (unsigned char)*mem);
	} else if (bytes==2) {
	  if (*(unsigned short*)mem != 0xFFFF) {
	    ppsOK = 0;
	    sprintf(msg, "0x%04x", (unsigned short)*mem);
	  }
	} else if (bytes==4) {
	  if (*(unsigned long*)mem != 0xFFFFFFFF) {
	    ppsOK = 0;
	    sprintf(msg, "0x%04lx", (unsigned long)*mem);
	  }
	} else {
	  fprintf(stderr, "Unsuported number of bits (%d)\n", bytes*8);
	  exit(1);
	}
	if (!ppsOK) {
	  //printf("Missing PPS at offset 0x%lx (%s)\n", offset, msg);
	  printf("Missing PPS-- at offset %ld (%s)\n", offset, msg);
	  if (fail) return(2);
	}
      } else if (mode==CHECK_STATS) {
	/* Check sampler statistics. If bytes=1 just accumulate all
	   possible bit patterns. Otherwise we need to accumulate each
	   second byte seperately.  This will be interpreted at the
	   end of the file. Using a long we can accumulate a maximum
	   of 4 GB data (actually 10 with good statistics. 
	   The Huygens mode (8x16 MHz bands) and the BG3 mode (2 x 64 MHz)
	   needs to be done slightly differently.
 */
	if (offset+(unsigned int)nread>=UINT_MAX-1) { /* 4 samples/byte */
	  fprintf(stderr, "File too large (> 4 GB) for stat counters\n");
	  return(1);
	}
	ptr = mem;
	if (bytes==1) {
	  for (i=0; i<nread; i++) {
	    stats[*ptr]++;
	    ptr++;
	  }
	} else if (bytes==2) {
	  for (i=0; i<nread; i++) {
	    stats[*ptr +(i%2)*256]++;
	    ptr++;
	  }

	} else {
	  fprintf(stderr, "Unknown number of bytes\n");
	  return(1);
	}
	offset += nread;

      } else if (mode==SEARCH_PATTERN) {
	int n;
	unsigned char *mptr, *cptr;
	mptr = mem;
	n = nread;

	/* First check if there was a match crossing two buffer boundaries */
	if (poverlap>0) {
	  if (memcmp(mptr, pattern+poverlap, psize-poverlap)==0) {
	    printf("Found match at pos %ld (0x%lx)\n", 
		   offset-poverlap+headsize, offset-poverlap+headsize);
	    mptr += psize-poverlap;
	  } 
	  poverlap=0;
	}

	while (1) {
	  /* First search for the first character */
	  cptr = memchr(mptr, pattern[0], n);
	  if (cptr==NULL) {
	    break; /* Nothing to see here */
	  }
	  if (cptr+psize>mptr+n) { 
	    /* Found at end of buffer, need special handling */

	    if (memcmp(cptr, pattern, mptr+n-cptr)==0) {  
	      /* A match at the end */
	      poverlap = mptr+n-cptr;
	      break;
	    }
	  } else {
	    if (memcmp(cptr, pattern, psize)==0) { /* They are the same */
	      printf("Found match at pos %ld (%lx)!!\n", 
		     (cptr-mem)+offset+headsize, (cptr-mem)+offset+headsize);
	      cptr+= psize-1;
	    }
	  }
	  
	  mptr = cptr+1;
	  n = nread - (mptr-mem);
	}

	offset += nread;
      } else {
	fprintf(stderr, "Internal error 2\n");
	return(1);
      }
      if (maxbytes>0 && offset>maxbytes) break;
    }
    status = close(file);

    if (mode==CHECK_STATS) {
      unsigned long avstats[8][4], totalcounts;
      int decode[4];
      char bits[4][3] = {"00", "01", "10", "11"};

      if (encoding==AT) {
	decode[0] = 3;
	decode[1] = 1;
	decode[2] = 0;
	decode[3] = 2;
      } else { /* Must be VLBA */
	decode[0] = 3;
	decode[1] = 2;
	decode[2] = 1;
	decode[3] = 0;
      }

      for (i=0; i<8; i++) {
	for (j=0; j<4; j++) {
	  avstats[i][j] = 0;
	}
      }
      totalcounts = 0;

      /* Need to report. Maybe change this to every second */

	printf("         high-  low-   low+   high+\n");
	printf("      ");
	for (j=0; j<4; j++) {
	  printf("     %s", bits[decode[j]]);
	}
	printf("\n");

      if (bytes==1) {
	for (i=0; i<256; i++) { /* Go through each byte distibuton */
	  for (j=0; j<4; j++) { /* Do each 2 bit sample seperately */
	    avstats[j][i>>(j*2)&0x3] += stats[i];
	  }
	  totalcounts += stats[i];
	}

	for (i=0; i<4; i++) {
	  printf("Chan %d:", i);
	  for (j=0; j<4; j++) {
	    printf(" %6.2f", avstats[i][decode[j]]/(float)totalcounts*100);
	  }
	  printf("\n");
	}

      } else {
	if (mhz64) {	    // 64 MHz version 
	  for (i=0; i<256; i++) { /* Go through each byte distribution */
	    for (j=0; j<4; j++) { /* Do each 2 bit sample seperately */
	      avstats[0][i>>(j*2)&0x3] += stats[i];
	      avstats[1][i>>(j*2)&0x3] += stats[256+i];
	    }
	    totalcounts += 4*stats[i]+ 4*stats[256+i]; /* Must be same total number */
	  }
	  totalcounts /= 2; /* We have counted twice */

	  for (i=0; i<2; i++) {
	    printf("Chan %d:", i);
	    for (j=0; j<4; j++) {
	      printf(" %6.2f", avstats[i][decode[j]]/(float)totalcounts*100);
	    }
	    printf("\n");
	  }
	} else { /* Huygens mode */

	  for (i=0; i<256; i++) { /* Go through each byte distibuton */
	    for (j=0; j<4; j++) { /* Do each 2 bit sample seperately */
	      avstats[j][i>>(j*2)&0x3] += stats[i];
	      avstats[j+4][i>>(j*2)&0x3] += stats[256+i];
	    }
	    totalcounts += stats[i] + stats[256+i];
	  }
	  totalcounts /= 2; /* We have counted twice */
	
	  for (i=0; i<8; i++) {
	    printf("Chan %d:", i);
	    for (j=0; j<4; j++) {
	      printf(" %6.2f", avstats[i][decode[j]]/(float)totalcounts*100);
	    }
	    printf("\n");
	  }
	}
      }
      printf("Total=%ld\n", totalcounts);
    }
  }
  return(1);
}

  
